@@ -28,11 +28,6 @@
 #'
 #' @keywords internal
 getGlobalsAndPackages <- function(expr, envir = parent.frame(), tweak = tweakExpression, globals = TRUE, resolve = getOption("future.globals.resolve", FALSE), persistent = FALSE, maxSize = getOption("future.globals.maxSize", 500 * 1024 ^ 2), ...) {
-  ## Nothing to do?
-  if (is.logical(globals) && !globals) {
-    return(list(expr = expr, globals = list(), packages = character(0)))
-  }
-  
   debug <- getOption("future.debug", FALSE)
   if (debug) mdebug("getGlobalsAndPackages() ...")
   
@@ -50,26 +45,68 @@ getGlobalsAndPackages <- function(expr, envir = parent.frame(), tweak = tweakExp
     mustExist <- is.element(globals.onMissing, "error")
   }
 
-
-  ## Alt 1. Identify globals based on expr, envir and tweak
   if (is.logical(globals)) {
     stop_if_not(length(globals) == 1, !is.na(globals))
-    if (debug) mdebug("Searching for globals ...")
-    ## Algorithm for identifying globals
-    globals.method <- getOption("future.globals.method", "ordered")
-    globals <- globalsOf(
-                 ## Passed to globals::findGlobals()
-                 expr, envir = envir, substitute = FALSE, tweak = tweak,
-                 ## Passed to globals::findGlobals() via '...'
-                 dotdotdot = "return",
-                 method = globals.method,
-                 unlist = TRUE,
-                 ## Passed to globals::globalsByName()
-                 mustExist = mustExist,
-                 recursive = TRUE
-               )
-    if (debug) mdebug("- globals found: [%d] %s", length(globals), hpaste(sQuote(names(globals))))
-    if (debug) mdebug("Searching for globals ... DONE")
+
+    ## Any manually added globals?
+    add <- attr(globals, "add", exact = TRUE)
+    if (!is.null(add)) {
+      if (is.character(add)) {
+        if (debug) mdebug("Retrieving 'add' globals ...")
+        add <- globalsByName(add, envir = envir, mustExist = mustExist)
+        if (debug) mdebug("- 'add' globals retrieved: [%d] %s", length(add), hpaste(sQuote(names(add))))
+        if (debug) mdebug("Retrieving 'add' globals ... DONE")
+      } else if (inherits(add, "Globals")) {
+        if (debug) mdebug("- 'add' globals passed as-is: [%d] %s", length(add), hpaste(sQuote(names(add))))
+      } else if (is.list(add)) {
+        if (debug) mdebug("- 'add' globals passed as-list: [%d] %s", length(add), hpaste(sQuote(names(add))))
+      } else {
+        stop("Attribute 'add' of argument 'globals' must be either a character vector or a named list: ", mode(add))
+      }
+      add <- as.FutureGlobals(add)
+      stop_if_not(inherits(add, "FutureGlobals"))
+    }
+    
+    ## Any manually dropped/ignored globals?
+    ignore <- attr(globals, "ignore", exact = TRUE)
+    if (!is.null(ignore)) {
+      stop_if_not(is.character(ignore))
+    }
+  
+    if (globals) {
+      if (debug) mdebug("Searching for globals...")
+      ## Algorithm for identifying globals
+      globals.method <- getOption("future.globals.method", "ordered")
+      globals <- globalsOf(
+                   ## Passed to globals::findGlobals()
+                   expr, envir = envir, substitute = FALSE, tweak = tweak,
+                   ## Passed to globals::findGlobals() via '...'
+                   dotdotdot = "return",
+                   method = globals.method,
+                   unlist = TRUE,
+                   ## Passed to globals::globalsByName()
+                   mustExist = mustExist,
+                   recursive = TRUE
+                 )
+      if (debug) mdebug("- globals found: [%d] %s", length(globals), hpaste(sQuote(names(globals))))
+      if (debug) mdebug("Searching for globals ... DONE")
+    } else {
+      if (debug) mdebug("Not searching for globals")
+      globals <- FutureGlobals()
+    }
+
+    ## Drop 'ignore' globals?
+    ## FIXME: This should really be implemented in globals::globalsOf()
+    if (!is.null(ignore)) {
+      if (any(ignore %in% names(globals))) {
+        globals <- globals[setdiff(names(globals), ignore)]
+      }
+    }
+  
+    ## Append 'add' globals?
+    if (inherits(add, "FutureGlobals")) {
+      globals <- unique(c(globals, add))
+    }
   } else if (is.character(globals)) {
     if (debug) mdebug("Retrieving globals ...")
     globals <- globalsByName(globals, envir = envir, mustExist = mustExist)
@@ -96,7 +133,7 @@ getGlobalsAndPackages <- function(expr, envir = parent.frame(), tweak = tweakExp
   }
 
   ## Are globals already resolved?
-  t <- attr(globals, "resolved")
+  t <- attr(globals, "resolved", exact = TRUE)
   if (isTRUE(t)) {
     resolve <- FALSE
     if (debug) mdebug("Resolving globals: %s (because already done)", resolve)
@@ -141,7 +178,7 @@ getGlobalsAndPackages <- function(expr, envir = parent.frame(), tweak = tweakExp
     globals <- as.FutureGlobals(globals)
 
     ## Unless already resolved, perform a shallow resolve
-    if (attr(globals, "resolved")) {
+    if (attr(globals, "resolved", exact = TRUE)) {
       idxs <- which(unlist(lapply(globals, FUN = inherits, "Future"), use.names = FALSE))
       if (debug) mdebug("Number of global futures: %d", length(idxs))
       
@@ -174,7 +211,7 @@ getGlobalsAndPackages <- function(expr, envir = parent.frame(), tweak = tweakExp
     ## Drop all globals which are located in one of
     ## the packages in 'pkgs'.  They will be available
     ## since those packages are attached.
-    where <- attr(globals, "where")
+    where <- attr(globals, "where", exact = TRUE)
 
     names <- names(globals)
     keep <- rep(TRUE, times = length(globals))
