@@ -205,7 +205,7 @@ for (type in types) {
   message("*** cluster() - assert registry behavior ...")
   
   ## Explicitly created clusters are *not* added to the registry
-  cl <- parallel::makeCluster(cores, type = type)
+  cl <- parallel::makeCluster(1L, type = type)
   plan(cluster, workers = cl)
   clR <- ClusterRegistry("get")
   stopifnot(is.null(clR))
@@ -228,11 +228,17 @@ for (type in types) {
   str(cl)
   parallel::stopCluster(cl)
   plan(sequential)
-  
+
+  cl1 <- parallel::makeCluster(1L, type = type)
+  plan(cluster, workers = cl1)
+  f1 <- future(1)
+
+  cl1 <- parallel::makeCluster(1L, type = type)
+  plan(cluster, workers = cl1)
+
   message(sprintf("Test set #1 with cluster type %s ... DONE", sQuote(type)))
 } ## for (type ...)
 
-
 library("parallel")
 for (type in types) {
   if (on_solaris) next
@@ -288,7 +294,7 @@ for (type in types) {
   if (type != "FORK" || getRversion() >= "3.3.0") {
   message("*** cluster() - crashed worker ...")
   
-  plan(cluster, workers = cl)
+  plan(cluster, workers = cl, .skip = FALSE)
   x %<-% 42L
   stopifnot(x == 42L)
   
@@ -309,9 +315,9 @@ for (type in types) {
   ## Verify that the reset worked
   cl <- parallel::makeCluster(1L, type = type)
   print(cl)
-  plan(cluster, workers = cl)
-  x %<-% 42L
-  stopifnot(x == 42L)
+  plan(cluster, workers = cl, .skip = FALSE)
+  x %<-% 43L
+  stopifnot(x == 43L)
   
   message("*** cluster() - crashed worker ... DONE")
   } ## if (type != "FORK" || getRversion() >= "3.2.0")
@@ -330,6 +336,24 @@ for (type in types) {
   message(sprintf("Test set #3 with cluster type %s ... DONE", sQuote(type)))
 } ## for (type ...)
 
+
+message("*** cluster() - exception when re-creating workers ...")
+## https://github.com/HenrikBengtsson/future/issues/261
+
+plan(cluster, workers = "localhost")
+f <- future(1)
+plan(cluster, workers = "localhost", .skip = FALSE)
+res <- tryCatch({
+  value(f)
+}, FutureError = function(ex) {
+  message(conditionMessage(ex))
+  ex
+})
+stopifnot(inherits(res, "FutureError"))
+
+message("*** cluster() - exception when re-creating workers ... DONE")
+
+
 message("*** cluster() ... DONE")
 
 ## Sanity checks
